<?php

namespace QUI\NodeDev;

use DirectoryIterator;

/**
 * Generates an import map for the given node_modules folder
 *
 * tried to implement: https://nodejs.org/docs/latest-v16.x/api/esm.html#esm_resolution_algorithm
 */
class ImportMapGenerator
{
    private string $nodeModulesPath = '';
    private string $baseDir = '';

    public function __construct(string $baseDir)
    {
        $realBaseDir = realpath($baseDir);

        if ($realBaseDir !== false) {
            $this->baseDir = $realBaseDir;
        }

        $this->nodeModulesPath = $this->baseDir . '/node_modules';
    }

    /**
     * Generates an import map for the given node_modules folder
     *
     * Tries to implement the resolution algorithm as described here:
     * https://nodejs.org/docs/latest-v16.x/api/esm.html#esm_resolution_algorithm
     *
     * Also adds a special case for Vue.js, to ensure that the correct file is loaded.
     *
     * @return array The generated import map
     */
    public function generateImportMap(): array
    {
        $importMap = ['imports' => []];
        $packages = $this->getPackages($this->nodeModulesPath);

        foreach ($packages as $packageDir) {
            $packageJsonPath = $packageDir . '/package.json';
            $packageJson = false;

            if (!file_exists($packageJsonPath)) {
                continue;
            }

            $packageJsonContent = file_get_contents($packageJsonPath);

            if ($packageJsonContent !== false) {
                $packageJson = json_decode($packageJsonContent, true);
            }

            if (!$packageJson || !isset($packageJson['name'])) {
                continue;
            }

            // Feste Regel für Vue.js, um sicherzustellen, dass die richtige Datei geladen wird
            if ($packageJson['name'] === 'vue') {
                $importMap['imports']['vue'] = './node_modules/vue/dist/vue.esm-browser.js';
                continue;
            }

            $packageEntries = $this->getPackageEntryPoints($packageDir, $packageJson);

            foreach ($packageEntries as $entryName => $packagePath) {
                $importMap['imports'][$entryName] = $packagePath;
            }
        }

        return $importMap;
    }

    /**
     * Scans the given node_modules directory and returns all packages
     *
     * @param string $nodeModulesPath The path to the node_modules directory
     * @return array A list of all packages
     */
    private function getPackages(string $nodeModulesPath): array
    {
        $packages = [];
        $this->scanPackages($nodeModulesPath, $packages);
        return $packages;
    }

    /**
     * Scans the given directory and adds all packages to the given array.
     * This function is recursive and will scan all subdirectories.
     *
     * @param string $directory The path to the directory to scan
     * @param array $packages The array to which the packages should be added
     */
    private function scanPackages(string $directory, array &$packages): void
    {
        $dirs = new DirectoryIterator($directory);

        foreach ($dirs as $dir) {
            if ($dir->isDot()) {
                continue;
            }

            if ($dir->isDir()) {
                $dirName = $dir->getFilename();

                if ($dirName === 'node_modules') {
                    continue;
                }

                $packageJsonPath = $dir->getPathname() . '/package.json';

                if (file_exists($packageJsonPath)) {
                    $packages[] = $dir->getPathname();
                } else {
                    $this->scanPackages($dir->getPathname(), $packages);
                }
            }
        }
    }

    /**
     * Returns an array of entry points for the given package. The array will have the following structure:
     * [
     *     'package-name' => 'path/to/module.js',
     *     'package-name/export-key' => 'path/to/export-key.js',
     * ]
     *
     * The function will first check if the package has exports. If it does, it will resolve the exports and add them to the array.
     * If the package does not have exports, it will check for the 'module' key in the package.json. If it is set, the function will resolve the path and add it to the array.
     * If the package does not have exports or a 'module' key, the function will check for the 'main' key in the package.json. If it is set, the function will resolve the path and add it to the array.
     * If the package does not have exports, a 'module' key or a 'main' key, the function will add the path to the package's 'index.js' file.
     *
     * @param string $packageDir The path to the package directory
     * @param array $packageJson The contents of the package.json file
     * @return array An array of entry points for the package
     */
    private function getPackageEntryPoints(string $packageDir, array $packageJson): array
    {
        $entries = [];

        $packageName = $packageJson['name'];

        if (isset($packageJson['exports'])) {
            $exports = $packageJson['exports'];

            $resolvedExports = $this->resolveExports($exports, $packageDir);

            foreach ($resolvedExports as $exportKey => $exportPath) {
                $importName = $exportKey === '.' || $exportKey === '' ?
                    $packageName :
                    $packageName . '/' . ltrim($exportKey, './');
                $entries[$importName] = $exportPath;
            }

            return $entries;
        }

        if (isset($packageJson['module']) && is_string($packageJson['module'])) {
            $path = $this->resolvePath($packageDir, $packageJson['module']);

            if ($path) {
                $entries[$packageName] = $path;
            }

            return $entries;
        }

        if (isset($packageJson['main']) && is_string($packageJson['main'])) {
            $path = $this->resolvePath($packageDir, $packageJson['main']);

            if ($path) {
                $entries[$packageName] = $path;
            }

            return $entries;
        }

        if (file_exists($packageDir . '/index.js')) {
            $path = './' . $this->getRelativePath($this->baseDir, $packageDir . '/index.js');
            $entries[$packageName] = $path;
        }

        return $entries;
    }

    /**
     * Resolve the exports section of a package.json to an array of import maps
     *
     * @param array|string $exports
     * @param string $packageDir
     * @param string $exportKey
     * @return array
     */
    private function resolveExports(array|string $exports, string $packageDir, string $exportKey = '.'): array
    {
        $entries = [];

        if (is_string($exports)) {
            $path = $this->resolvePath($packageDir, $exports);
            if ($path) {
                $entries[$exportKey] = $path;
            }
        } elseif (is_array($exports)) {
            // Priorisierung der Bedingungen
            $conditionsOrder = ['browser', 'development', 'default'];

            foreach ($conditionsOrder as $condition) {
                if (isset($exports[$condition])) {
                    $value = $exports[$condition];
                    $resolvedValue = $this->resolveExports($value, $packageDir, $exportKey);
                    $entries = array_merge($entries, $resolvedValue);
                    break;
                }
            }

            // Falls keine der priorisierten Bedingungen zutrifft
            if (empty($entries)) {
                foreach ($exports as $key => $value) {
                    if ($key === 'require') {
                        continue;
                    }
                    if ($key === 'import' || $key === 'default') {
                        $resolvedValue = $this->resolveExports($value, $packageDir, $exportKey);
                        $entries = array_merge($entries, $resolvedValue);
                        break;
                    }
                    if (str_starts_with($key, './')) {
                        $subExportKey = $key;
                        $resolvedValue = $this->resolveExports($value, $packageDir, $subExportKey);
                        $entries = array_merge($entries, $resolvedValue);
                    }
                }
            }
        }

        return $entries;
    }

    /**
     * Resolve a path relative to a package directory to an absolute path.
     *
     * If the path does not exist, the function returns null.
     *
     * @param string $packageDir The package directory
     * @param string $path The path to resolve, relative to the package directory
     *
     * @return string|null The absolute path, or null if it does not exist
     */
    private function resolvePath(string $packageDir, string $path): ?string
    {
        $resolvedPath = realpath($packageDir . '/' . ltrim($path, './'));

        if (!$resolvedPath) {
            return null;
        }

        $relativePath = './' . $this->getRelativePath($this->baseDir, $resolvedPath);
        return str_replace('\\', '/', $relativePath);
    }

    /**
     * Returns the relative path from $from to $to.
     *
     * This function takes two absolute paths, $from and $to, and returns the
     * relative path from $from to $to. If either of the paths does not exist,
     * an empty string is returned.
     *
     * @param string $from The from path
     * @param string $to The to path
     *
     * @return string The relative path
     */
    private function getRelativePath(string $from, string $to): string
    {
        $fromReal = realpath($from);
        $toReal = realpath($to);

        if (!$fromReal || !$toReal) {
            return '';
        }

        $from = str_replace('\\', '/', $fromReal);
        $to = str_replace('\\', '/', $toReal);
        $fromParts = explode('/', $from);
        $toParts = explode('/', $to);
        $length = min(count($fromParts), count($toParts));
        $commonLength = 0;

        for (; $commonLength < $length; $commonLength++) {
            if ($fromParts[$commonLength] !== $toParts[$commonLength]) {
                break;
            }
        }

        $upLevels = array_fill(0, count($fromParts) - $commonLength, '..');
        $remainingParts = array_slice($toParts, $commonLength);

        return implode('/', array_merge($upLevels, $remainingParts));
    }
}
